@@ -5,6 +5,9 @@ import requests
 import logging
 
 from requests import Response
+from requests.adapters import HTTPAdapter
+from urllib3.util.retry import Retry
+
 import aleph_alpha_client
 from aleph_alpha_client.document import Document
 from aleph_alpha_client.embedding import SemanticEmbeddingRequest
@@ -30,6 +33,17 @@ class AlephAlphaClient:
 
         self.request_timeout_seconds = request_timeout_seconds
 
+        retry_strategy = Retry(
+            total=3,
+            backoff_factor=0.1,
+            status_forcelist=[408, 429, 500, 502, 503, 504],
+            allowed_methods=["POST", "GET"],
+        )
+        adapter = HTTPAdapter(max_retries=retry_strategy)
+        self.requests_session = requests.Session()
+        self.requests_session.mount("https://", adapter)
+        self.requests_session.mount("http://", adapter)
+
         # check server version
         expect_release = "1"
         version = self.get_version()
@@ -57,10 +71,12 @@ class AlephAlphaClient:
             raise ValueError("cannot get token")
 
     def get_request(self, url, headers=None):
-        return requests.get(url, headers=headers, timeout=self.request_timeout_seconds)
+        return self.requests_session.get(
+            url, headers=headers, timeout=self.request_timeout_seconds
+        )
 
     def post_request(self, url, json, headers=None):
-        return requests.post(
+        return self.requests_session.post(
             url, headers=headers, json=json, timeout=self.request_timeout_seconds
         )
 
@@ -115,7 +131,7 @@ class AlephAlphaClient:
         self,
         model: str,
         prompt: Union[str, List[Union[str, ImagePrompt]]] = "",
-        hosting: str = "cloud",
+        hosting: Optional[str] = None,
         maximum_tokens: Optional[int] = 64,
         temperature: Optional[float] = 0.0,
         top_k: Optional[int] = 0,
@@ -126,6 +142,7 @@ class AlephAlphaClient:
         use_multiplicative_presence_penalty: Optional[bool] = False,
         penalty_bias: Optional[str] = None,
         penalty_exceptions: Optional[List[str]] = None,
+        penalty_exceptions_include_stop_sequences: Optional[bool] = None,
         best_of: Optional[int] = None,
         n: Optional[int] = 1,
         logit_bias: Optional[Dict[int, float]] = None,
@@ -144,8 +161,8 @@ class AlephAlphaClient:
             prompt (str, optional, default ""):
                 The text to be completed. Unconditional completion can be started with an empty string (default). The prompt may contain a zero shot or few shot task.
 
-            hosting (str, optional, default "cloud"):
-                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+            hosting (str, optional, default None):
+                Specifies where the computation will take place. This defaults to None, meaning that it can be
                 executed on any of our servers. An error will be returned if the specified hosting is not available.
                 Check available_models() for available hostings.
 
@@ -177,6 +194,40 @@ class AlephAlphaClient:
             use_multiplicative_presence_penalty (bool, optional, default True)
                 Flag deciding whether presence penalty is applied multiplicatively (True) or additively (False). This changes the formula stated for presence and frequency penalty.
 
+            penalty_bias (string, optional)
+                If set, all tokens in this text will be used in addition to the already penalized tokens for repetition penalties. These consist of the already generated completion tokens and the prompt tokens, if `repetition_penalties_include_prompt` is set to `true`.
+
+                *Potential use case for a chatbot-based completion:*
+
+                Instead of using `repetition_penalties_include_prompt`, construct a new string with only the chatbot's reponses included. You would leave out any tokens you use for stop sequences (i.e. `\nChatbot:`), and all user messages.
+
+                With this bias, if you turn up the repetition penalties, you can avoid having your chatbot repeat itself, but not penalize the chatbot from mirroring language provided by the user.
+
+            penalty_exceptions (List(str), optional)
+                List of strings that may be generated without penalty, regardless of other penalty settings.
+
+                This is particularly useful for any completion that uses a structured few-shot prompt. For example, if you have a prompt such as:
+
+                ```
+                I want to travel to a location, where I can enjoy both beaches and mountains.
+
+                - Lake Garda, Italy. This large Italian lake in the southern alps features gravel beaches and mountainside hiking trails.
+                - Mallorca, Spain. This island is famous for its sandy beaches, turquoise water and hilly landscape.
+                - Lake Tahoe, California. This famous lake in the Sierra Nevada mountains offers an amazing variety of outdoor activities.
+                -
+                ```
+
+                You could set `penalty_exceptions` to `["\n-"]` to not penalize the generation of a new list item, but still increase other penalty settings to encourage the generation of new list items without repeating itself.
+
+                By default, we will also include any `stop_sequences` you have set, since completion performance can be degraded if expected stop sequences are penalized. You can disable this behavior by settings `penalty_exceptions_include_stop_sequences` to `false`.
+
+            penalty_exceptions_include_stop_sequences (bool, optional, default true)
+                By default, we include any `stop_sequences` in `penalty_exceptions`, to not penalize the presence of stop sequences that are present in few-shot prompts to provide structure to your completions.
+
+                You can set this to `false` if you do not want this behavior.
+
+                See the description of `penalty_exceptions` above for more information on what `penalty_exceptions` are used for.
+
             best_of (int, optional, default None)
                 Generates best_of completions server-side and returns the "best" (the one with the highest log probability per token). Results cannot be streamed.
                 When used with n, best_of controls the number of candidate completions and n specifies how many to return – best_of must be greater than n.
@@ -220,7 +271,6 @@ class AlephAlphaClient:
         payload = {
             "model": model,
             "prompt": _to_serializable_prompt(prompt=prompt),
-            "hosting": hosting,
             "maximum_tokens": maximum_tokens,
             "temperature": temperature,
             "top_k": top_k,
@@ -229,6 +279,7 @@ class AlephAlphaClient:
             "frequency_penalty": frequency_penalty,
             "penalty_bias": penalty_bias,
             "penalty_exceptions": penalty_exceptions,
+            "penalty_exceptions_include_stop_sequences": penalty_exceptions_include_stop_sequences,
             "best_of": best_of,
             "n": n,
             "logit_bias": logit_bias,
@@ -240,6 +291,9 @@ class AlephAlphaClient:
             "disable_optimizations": disable_optimizations,
         }
 
+        if hosting is not None:
+            payload["hosting"] = hosting
+
         response = self.post_request(
             self.host + "complete",
             headers=self.request_headers,
@@ -259,7 +313,7 @@ class AlephAlphaClient:
         prompt: Union[str, Sequence[Union[str, ImagePrompt]]],
         pooling: List[str],
         layers: List[int],
-        hosting: str = "cloud",
+        hosting: Optional[str] = None,
         tokens: Optional[bool] = False,
         type: Optional[str] = None,
     ):
@@ -287,8 +341,8 @@ class AlephAlphaClient:
                     * last_token: just use the last token
                     * abs_max: aggregate token embeddings across the sequence dimension using a maximum of absolute values
 
-            hosting (str, optional, default "cloud"):
-                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+            hosting (str, optional, default None):
+                Specifies where the computation will take place. This defaults to None, meaning that it can be
                 executed on any of our servers. An error will be returned if the specified hosting is not available.
                 Check available_models() for available hostings.
 
@@ -310,12 +364,15 @@ class AlephAlphaClient:
         payload = {
             "model": model,
             "prompt": serializable_prompt,
-            "hosting": hosting,
             "layers": layers,
             "tokens": tokens,
             "pooling": pooling,
             "type": type,
         }
+
+        if hosting is not None:
+            payload["hosting"] = hosting
+
         response = self.post_request(
             self.host + "embed", headers=self.request_headers, json=payload
         )
@@ -325,7 +382,7 @@ class AlephAlphaClient:
         self,
         model: str,
         request: SemanticEmbeddingRequest,
-        hosting: str = "cloud",
+        hosting: Optional[str] = None,
     ):
         """
         Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).
@@ -335,7 +392,7 @@ class AlephAlphaClient:
                 Name of model to use. A model name refers to a model architecture (number of parameters among others). Always the latest version of model is used. The model output contains information as to the model version.
 
             hosting (str, required):
-                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+                Specifies where the computation will take place. This defaults to None, meaning that it can be
                 executed on any of our servers. An error will be returned if the specified hosting is not available.
                 Check available_models() for available hostings.
 
@@ -349,11 +406,14 @@ class AlephAlphaClient:
 
         payload: Dict[str, Any] = {
             "model": model,
-            "hosting": hosting,
             "prompt": serializable_prompt,
             "representation": request.representation.value,
             "compress_to_size": request.compress_to_size,
         }
+
+        if hosting is not None:
+            payload["hosting"] = hosting
+
         response = self.post_request(
             self.host + "semantic_embed", headers=self.request_headers, json=payload
         )
@@ -363,7 +423,7 @@ class AlephAlphaClient:
         self,
         model,
         completion_expected,
-        hosting: str = "cloud",
+        hosting: Optional[str] = None,
         prompt: Union[str, List[Union[str, ImagePrompt]]] = "",
     ):
         """
@@ -376,8 +436,8 @@ class AlephAlphaClient:
             completion_expected (str, required):
                 The ground truth completion expected to be produced given the prompt.
 
-            hosting (str, optional, default "cloud"):
-                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+            hosting (str, optional, default None):
+                Specifies where the computation will take place. This defaults to None, meaning that it can be
                 executed on any of our servers. An error will be returned if the specified hosting is not available.
                 Check available_models() for available hostings.
 
@@ -390,9 +450,12 @@ class AlephAlphaClient:
         payload = {
             "model": model,
             "prompt": serializable_prompt,
-            "hosting": hosting,
             "completion_expected": completion_expected,
         }
+
+        if hosting is not None:
+            payload["hosting"] = hosting
+
         response = self.post_request(
             self.host + "evaluate", headers=self.request_headers, json=payload
         )
@@ -408,7 +471,7 @@ class AlephAlphaClient:
         disable_optimizations: bool = False,
         max_answers: int = 0,
         min_score: float = 0.0,
-        hosting: str = "cloud",
+        hosting: Optional[str] = None,
     ):
         """
         Answers a question about a prompt.
@@ -447,8 +510,8 @@ class AlephAlphaClient:
             min_score (float, default 0.0):
                 The lower limit of minimum score for every answer.
 
-            hosting (str, default "cloud"):
-                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+            hosting (str, default None):
+                Specifies where the computation will take place. This defaults to None, meaning that it can be
                 executed on any of our servers. An error will be returned if the specified hosting is not available.
                 Check available_models() for available hostings.
         """
@@ -464,9 +527,11 @@ class AlephAlphaClient:
             "min_score": min_score,
             "max_chunk_size": max_chunk_size,
             "disable_optimizations": disable_optimizations,
-            "hosting": hosting,
         }
 
+        if hosting is not None:
+            payload["hosting"] = hosting
+
         response = self.post_request(
             self.host + "qa",
             headers=self.request_headers,
